use async_trait::async_trait;
use invidious::hidden::SearchItem;
use invidious::{ClientAsync, ClientAsyncTrait, MethodAsync};
use itertools::Itertools;
#[cfg(test)]
use mockall::{automock, predicate::*};
use std::error::Error;
use std::sync::Arc;

#[cfg_attr(test, automock)]
#[async_trait]
pub(crate) trait SearchableClient {
    async fn search_video<'a>(
        &self,
        id: &str,
        sort_by: Option<&'a SearchSortBy>,
    ) -> Result<VideoSearch, Box<dyn Error>>;
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub(crate) struct VideoSearch {
    pub(crate) items: Vec<Video>,
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub(crate) struct Video {
    pub(crate) title: String,
    pub(crate) video_id: String,
    pub(crate) author: String,
    pub(crate) author_id: String,
    pub(crate) author_url: String,
    pub(crate) length_seconds: u64,
    pub(crate) description: String,
    pub(crate) description_html: String,
    pub(crate) view_count: u64,
    pub(crate) published: u64,
    pub(crate) published_text: String,
    pub(crate) live_now: bool,
    pub(crate) premium: bool,
}

type SearchSortBy = str;

#[allow(dead_code)]
#[allow(unused_variables)]
const BY_RELEVANCE: &SearchSortBy = "relevance";
#[allow(dead_code)]
#[allow(unused_variables)]
const BY_RATING: &SearchSortBy = "rating";
#[allow(dead_code)]
#[allow(unused_variables)]
const BY_UPLOAD_DATE: &SearchSortBy = "upload_date";
#[allow(dead_code)]
#[allow(unused_variables)]
const BY_VIEW_COUNT: &SearchSortBy = "view_count";

#[derive(Clone)]
pub(crate) struct Client {
    client: Arc<ClientAsync>,
}

impl Client {
    pub(crate) async fn new() -> Self {
        // TODO check for a stable instance, or rotate between a pool of stable ones
        let client = ClientAsync::new(
            String::from("https://inv.bp.projectsegfau.lt"),
            MethodAsync::default(),
        );

        Client {
            client: Arc::new(client),
        }
    }

    #[allow(dead_code)]
    #[allow(unused_variables)]
    #[cfg(test)]
    pub(crate) fn new_with_dependencies(client: invidious::ClientAsync) -> Self {
        Client {
            client: Arc::new(client),
        }
    }
}

#[async_trait]
impl SearchableClient for Client {
    async fn search_video<'a>(
        &self,
        keyword: &str,
        sort_by: Option<&'a SearchSortBy>,
    ) -> Result<VideoSearch, Box<dyn Error>> {
        let mut query = Vec::<String>::new();
        query.push(format!("{}={}", "q", keyword));
        match sort_by {
            Some(sorting_type) => query.push(format!("{}={}", "sort_by", sorting_type)),
            None => {}
        }
        query.push(format!("{}={}", "type", "video"));

        let generated_query = query.join(",");
        let search = self.client.search(Some(generated_query.as_str())).await?;

        let videos: Vec<Video> = search
            .items
            .iter()
            .map(|search_result| {
                let video = match search_result {
                    SearchItem::Video(item) => Some(Video {
                        title: item.title.clone(),
                        video_id: item.id.clone(),
                        author: item.author.clone(),
                        author_id: item.author_id.clone(),
                        author_url: item.author_url.clone(),
                        length_seconds: item.length.clone() as u64,
                        description: item.description.clone(),
                        description_html: item.description_html.clone(),
                        view_count: item.views.clone(),
                        published: item.published.clone(),
                        published_text: item.published_text.clone(),
                        live_now: item.live.clone(),
                        premium: item.premium.clone(),
                    }),
                    _ => None,
                };
                video
            })
            .filter(|x| x.is_some())
            .map(|x| x.unwrap())
            .filter(|x| !x.premium && !x.live_now)
            .sorted_by(|a, b| b.view_count.cmp(&a.view_count))
            .collect();

        Result::Ok(VideoSearch { items: videos })
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn should_search_properly() {
        // FIXME: the real solution for properly testing this piece of code would be to spin up a Mocked Invidious
        // instance (i.e. mock-server) and mock the server responses. With the current implementation, the tests
        // will always be flaky.
        let client = tokio_test::block_on(Client::new());

        let result = tokio_test::block_on(
            client.search_video("Eminem - Lose Yourself", Some(BY_UPLOAD_DATE)),
        )
        .unwrap();
        assert_eq!(19, result.items.len());
        // Just check the first one for now - in the future we can check for the whole array
        assert_eq!("Eminem - Lose Yourself [HD]", result.items[0].title);
    }
}
